/* Catch.h          -*-objc-*-
 *
 * catch and throw for Objective-C programs.
 *
 * Copyright Niels Mller <nisse@lysator.liu.se> 1995
 *
 * Freely distributable under the terms and conditions of the
 * GNU General Public License.
 */

#ifndef CATCH_OBJC_H_INCLUDED
#define CATCH_OBJC_H_INCLUDED

#include <objects/StackFrame.h>
#include <objects/Exception.h>
#include <objects/Jumping.h>

@interface Catch_common : StackFrame <Jumping>
{
  JMP_BUF where;
}
- (JMP_BUF *) where;
- (JMP_BUF *) catch;
@end /* Catch_common */


@interface Catch : Catch_common
{
  id result;
}
- value;
- value: newValue;
- (void) throw: value;
- (void) throw: value free: (BOOL) freeFlag;
@end /* Catch */


@interface CatchException : Catch_common
{
  id exception;
  exception_type exceptionType;
}
- (JMP_BUF *) catch: type;
- exceptionType;
- exceptionType: type;
- exception;
- exception: object;
@end /* CatchException */


/* Macros */

/* For your information: I'm really ambivalent about how the
 * interface to exceptions should be designed. At last, I decided
 * to use macros with blocks as arguments, instead of the NextStep
 * style with NS* macros to start and end the blocks.
 *
 * One main reason is that I want as few macro names to invent and
 * remember as possible.
 */
   
/* CATCH
 *
 * Exceutes the block BODY, catching any throws to the tag TAG.
 * The value is 1 if a throw happened, otherwise 0. In the first case,
 * [TAG value] gives the value from the throw.
 *
 * Allocation and freeing of TAG, which should be a Catch object,
 * is not done by the CATCH macro.
 *
 * Usage
 * ~~~~~
 * tag = [[Catch alloc] init];
 * if (CATCH(tag,
 *          {
 *            ...
 *          }))
 *    printf("Caught throw, value = %s\n", [[tag value] something]);
 * else
 *    printf("Done\n");
 * [tag free];
 */
 
#define CATCH(tag, body) \
( SETJMP(*[(tag) catch]) ? ([(tag) value], 1) : ((body), 0) )

/* TRY
 *
 * Executes the BODY, catching TYPE exceptions. If an exception
 * happens, VAR is bound to the exception, and HANDLER is
 * executed. When finished with the exception, you should send
 * it the `finished' message.
 *
 * Usage:
 *
 * id error;
 * TRY({
 *        body
 *     },
 *     @protocol(IOError), error,
 *     {
 *        [error something];
 *        [error finished];
 *     });
 */

#define TRY(body, type, var, handler) do { \
  id _TRY_tag = [[CatchException alloc] init]; \
  if (SETJMP(*[_TRY_tag catch: type]) == 0) \
    {\
       body \
       [_TRY_tag free];\
    } \
  else \
    {\
      (var) = [_TRY_tag exception]; \
      [_TRY_tag free];\
      handler \
    } \
} while(0)
    

#endif CATCH_OBJC_H_INCLUDED
